@@ -4,7 +4,7 @@
     alea.py
 
 --------------------------------------------------------------------------------
-Copyright 2013-2020 Pierre Denis
+Copyright 2013-2021 Pierre Denis
 
 This file is part of Lea.
 
@@ -86,8 +86,9 @@ class Alea(Lea):
 
     # function used to simplify symbolic probability expressions to be
     # displayed (see _simplify and Alea.__init__ methods)
-    _symbolic_simplify_function = sympy and staticmethod(sympy.factor)
-
+    # setting it to None disables any symbolic simplification
+    _symbolic_simplify_function = sympy and staticmethod(sympy.simplify)
+    
     # dictionary used in _downcast method
     __downcast_class = dict( {Fraction : ExtFraction,
                               Decimal  : ExtDecimal })
@@ -150,7 +151,7 @@ class Alea(Lea):
             otherwise,
                 returns v unchanged
         '''
-        if sympy is not None and isinstance(v,sympy.Expr):
+        if Alea._symbolic_simplify_function is not None and isinstance(v,sympy.Expr):
             return Alea._symbolic_simplify_function(v)
         if to_float:
             return float(v)
@@ -254,14 +255,14 @@ class Alea(Lea):
             if nb_none > 1:
                 raise Lea.Error("for normalization, no more than one single probability can be None")
             if nb_none == 1:
-                p_sum = sum(p for p in ps if p is not None)
+                p_sum = Alea._simplify(sum(p for p in ps if p is not None))
                 Alea._check_prob(p_sum)
                 idx_none = ps.index(None)
                 ps = ps[:idx_none] + (1-p_sum,) + ps[idx_none+1:]
             else:
-                p_sum = sum(ps)
+                p_sum = Alea._simplify(sum(ps))
                 ps = (truediv(p,p_sum) for p in ps)
-            if sympy is not None and Alea._symbolic_simplify_function is not None and isinstance(p_sum,sympy.Expr):
+            if Alea._symbolic_simplify_function is not None and isinstance(p_sum,sympy.Expr):
                 ps = (Alea._symbolic_simplify_function(p) for p in ps)
         self._ps = tuple(ps)
         if len(self._vs) != len(self._ps):
@@ -315,7 +316,7 @@ class Alea(Lea):
         ## which is the highest arithmetic class in class hierarchy
         return Alea((value,),(1,),normalization=False,prob_type=prob_type)
 
-    def new(self,n=None,prob_type=-1,sorting=False):
+    def new(self,n=None,prob_type=-1,sorting=False,normalization=False):
         ''' returns a new Alea instance, which represents the same probability
             distribution as self but for another event, independent from the
             event represented by self;
@@ -329,12 +330,13 @@ class Alea(Lea):
                     (see doc of Alea.set_prob_type);
             * sorting allows sorting the value of the returned Alea instance
               (see Alea.pmf method);
+            * normalization (default: False): if True, then each probability is divided
+              by the sum of all probabilities              
             note that the present method overloads Lea.new to be more efficient;
         '''
         if sorting:
             # for sorting, relay to Lea.new, which is less efficient but handles the case correctly 
             return Lea.new(self,prob_type=prob_type,sorting=True)
-        normalization = prob_type != -1
         new_alea = Alea(self._vs,self._ps,normalization=normalization,prob_type=prob_type)
         if prob_type == -1:
             ## note that the new Alea instance shares the immutable _vs and _ps attributes of self
@@ -618,7 +620,7 @@ class Alea(Lea):
             to define non-uniform probability distribution, as alternative to repeating the
             same row multiple times
         '''
-        # Filter out kewords argument targeting Alea constructors (assuming no name collision csv.reader keywords) 
+        # filter out keywords argument targeting Alea constructors (assuming no name collision csv.reader keywords) 
         kwargs = dict((k,v) for (k,v) in fmtparams.items() if k in Alea.__contructor_arg_names)
         fmtparams = dict((k,v) for (k,v) in fmtparams.items() if k not in kwargs)
         if isinstance(csv_file,str):
@@ -856,7 +858,7 @@ class Alea(Lea):
             kind = '/'
         else:
             kind = None
-        return self.as_string(kind)
+        return self.get_alea().as_string(kind)
 
     __repr__ = __str__
           
@@ -982,12 +984,6 @@ class Alea(Lea):
         '''
         return collections.OrderedDict((v,Alea._downcast_prob(p)) for (v,p) in zip(self._vs,self.cumul()[1:]))
 
-    def get_alea_leaves_set(self):
-        ''' returns a set containing all the Alea leaves in the tree having the root self
-            in the present case of Alea instance, it returns the singleton set with self as element
-        '''
-        return frozenset((self,))        
-     
     def _get_lea_children(self):
         ''' see Lea._get_lea_children
         '''
@@ -1021,6 +1017,11 @@ class Alea(Lea):
                                      for (vx,px) in obs_pmf_tuple))
                              for v in self.support))
 
+    def is_bindable(self,v):
+        ''' see Lea.is_bindable
+        '''
+        return self._val is self and v in self._vs
+
     def observe(self,v):
         ''' (re)bind self with given value v;
             requires that self is an Alea instance (i.e. not dependent of other Lea instances);
@@ -1035,7 +1036,7 @@ class Alea(Lea):
             requires that self is an Alea instance (i.e. not dependent of other Lea instances);
             if check is True, then requires that self is bound
         '''
-        if check and self._val is self:
+        if check and not self.is_bound():
             raise Lea.Error("%s already unbound"%(self._id(),))
         self._val = self
 
@@ -1061,7 +1062,7 @@ class Alea(Lea):
         if p1 is None:
             # val is absent form self: the probability is null, casted in the type of the last probability found
             p1 = 0 * p
-        return Alea._simplify(p1)
+        return p1
 
     def cumul(self):
         ''' returns a list with the probabilities p that self <= value ;
@@ -1391,11 +1392,11 @@ class Alea(Lea):
                     res -= p*log2(p)
             return res
         except TypeError:
+            # sympy exception assumed: no ceiling
             try:
-                log = sympy.log
                 for (v,p) in self._gen_raw_vps():
-                    res -= p*log(p)
-                return res / log(2)
+                    res -= p*sympy.log(p)
+                return res / sympy.log(2)
             except:
                 raise Lea.Error("cannot calculate logarithm on given probability types")
 
@@ -1412,11 +1413,43 @@ class Alea(Lea):
         n = len(self._vs)
         if n == 1:
             return 0.0
-        res = self.entropy / log2(n)
+        entropy = self.entropy
         try:
-            return min(1.0,res)
+            return min(1.0,entropy/log2(n))
         except TypeError:
-            return res
+            # sympy exception assumed: no ceiling
+            return entropy / sympy.log(n,2)
+
+    def cross_entropy(self,lea1):
+        ''' static method, returns the cross-entropy between self and given lea1;
+            the logarithm base is 2;
+            requires that all values of lea1's support have a non-null probability in self;
+            notes:
+            - the cross-entropy is non-commutative;
+            - the cross-entropy should always be greater than the entropy of first argument,
+              the equality being reached if both arguments have same pmf; this is guaranteed
+              by the implementation, even in case of rounding errors;
+            - if self is interpreted as frequencies of observed data having N as total number
+              of samples, then the cross-entropy is linked to (negative) log-likelihood by
+                log-likelihood = - N * cross-entropy
+              using logarithm in base 2 (for other base, use the right factor)
+        '''
+        lea1_pmf_dict = Alea.coerce(lea1).pmf_dict
+        try:
+            ce = -sum(px*log2(lea1_pmf_dict[vx]) for (vx,px) in self._gen_vps() if px > 0)
+        except KeyError as key_error:
+            raise Lea.Error("observed value '%s' is not produced by given model"%(key_error.args[0],))
+        except ValueError:
+            raise Lea.Error("some observed value has null probability in given model")
+        except:
+            # sympy exception assumed due to log function or test px > 0:
+            # retry using sympy log funtion and without test clause
+            ce = -sum(px*sympy.log(lea1_pmf_dict[vx]) for (vx,px) in self._gen_vps()) / sympy.log(2)
+        try:        
+            return max(ce,self.entropy)
+        except TypeError:
+            # sympy exception assumed: no ceiling
+            return ce
 
     def redundancy(self):
         ''' returns the redundancy of self;
